# How to use:
# python graphs/genGraph.py data/morphisms.txt o 8 | dot -Tpdf -ographs/Orientable8.pdf
#
# tested with graphviz 2.36.0 (providing dot) 

import sys
import re

def read_morphisms(text):
    """
    Read a morphisms file of the form
    otet10_0004#0 -> otet05_0002#1
    """
    morphisms = []
    for line in text.split('\n'):
        stripped = line.strip()
        if stripped:
            src, trg = stripped.split('->')
            morphisms.append((src.strip(), trg.strip()))

    return morphisms

def filter_morphisms(morphisms, orientability, num_tets):
    result = []
    for morphism in morphisms:
        src, dst = morphism

        if orientability == 'n' or (
            get_orientability(src) == 'o' and
            get_orientability(dst) == 'o'):

            if (get_num_tet(src) <= num_tets and
                get_num_tet(dst) <= num_tets):

                result.append(morphism)
    return result

def split_triangulation_name(trig_name):
    """
    Splits otet10_0004#0 into ('otet10_0004', 0)
    """

    name, index_str = trig_name.split('#')
    return name, int(index_str)

def get_num_tet(name):
    """
    Given otet10_0004 or otet10_0004#0, returns 10.
    """

    return int(re.match('[on]tet0*(\d+)_0*\d+(#\d+)?', name).group(1))

def get_orientability(name):
    """
    Given otet10_0003, return 'o'.
    """
    return name[0]

def get_objects(morphisms):
    """
    Returns a set of all src and dsts in morphisms.
    """
    result = set()
    for src, dst in morphisms:
        result.add(src)
        result.add(dst)
    return result

def morphism_edges(morphisms):
    return ''.join(['    "%s" -> "%s";\n' % (src, trg)
                    for src, trg in morphisms])

def num_tets_spine(objects):
    num_tets_to_names = {}
    for object in objects:
        num_tets_to_names.setdefault(get_num_tet(object), []).append(object)

    min_tets = min(num_tets_to_names.keys())
    max_tets = max(num_tets_to_names.keys())

    spine = '    {\n        node [shape=box];\n        %s\n    }\n\n\n' % (
        ' -> '.join(['%d' % i for i in range(max_tets, min_tets-1, -1)]))

    ranks = ''
    for num in sorted(num_tets_to_names.keys()):
        names = num_tets_to_names[num]
        ranks += '    { rank = same; %d; %s; }\n' % (num, ';'.join(
                ['"%s"' % name for name in names]))

    return spine + ranks + '\n\n'

def isometry_arrows(objects):
    manifold_name_to_indices = {}
    for object in objects:
        manifold_name, index = split_triangulation_name(object)
        manifold_name_to_indices.setdefault(manifold_name, []).append(index)

    subgraphs = ''

    count = 0
    for manifold_name, indices in manifold_name_to_indices.items():
        if len(indices) > 1:

            subgraphs += '    subgraph cluster%d {\n        ' % count
            subgraphs += ' -> '.join(['"%s#%d"' % (manifold_name, index)
                                      for index in sorted(indices)])
            subgraphs += ' [dir=both, style=dashed];\n    }\n\n'
            count += 1
    return subgraphs

def digraph(morphisms):
    objects = get_objects(morphisms)

    return 'digraph g {\n\n%s\n\n%s\n\n%s\n}\n' % (
        num_tets_spine(objects),
        isometry_arrows(objects),
        morphism_edges(morphisms))

if __name__ == '__main__':
    if not len(sys.argv) == 4:
        print "Usage: python genGraph.py MORPHISMS_FILE [o|n] NUM_TETS"
        sys.exit(1)

    morphisms_file, orientability, num_tets_str = sys.argv[1:]
    num_tets = int(num_tets_str)

    all_morphisms = read_morphisms(open(morphisms_file).read())
    morphisms = filter_morphisms(all_morphisms, orientability, num_tets)

    print digraph(morphisms)
