"""
Verifying that a triangulation of a cusped manifold is canonical.

Relies on SnapPy and HIKMOT.

Typical usage to prove a triangulation is canonical.  

>>> import snappy, canonical
>>> M = snappy.Manifold('v3372')
>>> canonical.is_canonical_triangulation(M)
True

You can also examine things in more detail.  Notice
how the area is an interval.  

>>> C = canonical.CuspedManifold(M)
>>> C.cusp_area(0)
[8.55907009806665,8.5590700980696326]

Here is a provable upper bound on the maximum tilt; since
it's negative, Theorem 3.3 says this is the canonical triangulation.

>>> C.least_convexity()
-0.10058282733152665
>>> C.is_canonical_triangulation()
True

Can also be used to prove that a triangulation is *not* canonical.

>>> N = snappy.Manifold('5_2')
>>> D = canonical.CuspedManifold(N)
>>> D.is_canonical_triangulation()
False

As you can see, the largest tilt is an unambiguously postive
interval.  

>>> tilts = D.LHS_of_convexity_equations()
>>> max(tilts)
[0.12420920467412718,0.1242092046749759]

A manifold with 2 cusps.

>>> N = snappy.Manifold('m125')
>>> N.canonize()
>>> is_canonical_triangulation(N)
True

Here's an example where the canonical cellulation is a regular ideal
octahedron.   The tilt whose interval straddles 0 is really 0... 

>>> C = canonical.CuspedManifold('m137')
>>> C.is_canonical_triangulation()
'unknown'
>>> max(C.LHS_of_convexity_equations())
[-4.2188474935755959e-13,4.2255088317233469e-13]
"""
# This code is copyrighted by Nathan Dunfield, Neil Hoffman, and Joan Licata
# and released under the GNU GPL version 2 or (at your option) any later version.

# Simplified schema so that it can handle multiple cusps
#            - Matthias Goerner, 11/20/2014

import snappy, hikmot
import snappy.snap.t3mlite as t3m
from snappy.snap.t3mlite.simplex import *
import copy

#--- Some additional methods for t3m.Tetrahedron ----

def add_edge_parameters(self, shape):
    z = shape
    zp = 1/(1-z)
    zpp = (z-1)/z
    self.edge_params = {E01:z, E23:z, E02:zp, E13:zp, E03:zpp, E12:zpp}

def tilt(self, v):
    "The tilt of the face of the tetrahedron opposite the vertex v."
    ans = 0
    for w in ZeroSubsimplices:
        if v == w:
            c_w = 1
        else:
            z = self.edge_params[v | w]
            c_w = -z.real/abs(z)
        R_w = self.horotriangles[w].circumradius
        ans += c_w*R_w
    return ans

t3m.Tetrahedron.add_edge_parameters = add_edge_parameters
t3m.Tetrahedron.tilt = tilt

#--- Some additional methods for HIKMOT's interval types ----

def complex_abs(self):
    a, b = self.real, self.imag
    norm_squared = abs(a*a + b*b)
    return norm_squared.sqrt()

hikmot.complex.complex.__abs__ = complex_abs

def interval_lt(self, other):
    if not isinstance(other, self.__class__):
        other = self.__class__(other)
    return self.sup < other.inf

hikmot.interval.interval.__lt__ = interval_lt

def max_sup(interval_list):
    return max(i.sup for i in interval_list)

def max_inf(interval_list):
    return max(i.inf for i in interval_list)

#-------- t3m helper code --------

FacesAnticlockwiseAroundVertices = {
    V0 : (F1, F2, F3),
    V1 : (F0, F3, F2), 
    V2 : (F0, F1, F3),
    V3 : (F0, F2, F1)}

def glued_to(tetrahedron, face):
    """
    Returns (other tet, other face).
    """
    return tetrahedron.Neighbor[face], tetrahedron.Gluing[face].image(face)

def tets_and_vertices_of_cusp(cusp):
    return [(corner.Tetrahedron, corner.Subsimplex) for corner in cusp.Corners]

#-------- Main classes --------

class HoroTriangle:
    """
    A horosphere cross section in the corner of an ideal tetrahedron.
    The sides of the triangle correspond to faces of the tetrahedron.
    """
    def __init__(self, tet, vertex, known_side, length_of_side):
        sides = FacesAnticlockwiseAroundVertices[vertex]
        left, center, right = HoroTriangle._make_middle(sides, known_side)
        z_l = tet.edge_params[left & center]
        z_r = tet.edge_params[center & right]
        L = length_of_side
        self.lengths = {center:L, left:abs(z_l)*L, right:L/abs(z_r)}
        a, b, c = self.lengths.values()
        self.area =L*L*z_l.imag/2
        # Below is the usual formula for circumradius combined with Heron's formula.
        self.circumradius = a*b*c/(4*self.area)  

    def rescale(self, t):
        "Rescales the triangle by a Euclidean dilation"
        for face, length in self.lengths.items():
            self.lengths[face] = t*length
        self.circumradius = t*self.circumradius
        self.area = t*t*self.area

    @staticmethod
    def _make_middle((a,b,c), x):
        "Cyclically rotate (a,b,c) so that x is the middle entry"
        if x == a:
            return (c,a,b)
        elif x == b:
            return (a,b,c)
        elif x == c:
            return (b,c,a)


class CuspedManifold(t3m.Mcomplex):
    """
    A t3m triangulation built from a SnapPy manifold with
    a hyperbolic structure specified by HIKMOT's guaranteed
    shape intervals.
    """
    def __init__(self, manifold):
        if isinstance(manifold, str):
            manifold = snappy.Manifold(manifold)
        t3m.Mcomplex.__init__(self, manifold)
        self.snappy_manifold = manifold
        self.add_shapes()
        for T in self.Tetrahedra:
            T.horotriangles = {V0:None, V1:None, V2:None, V3:None}
        self.add_cusp_cross_sections()

        # Some sanity checks.  
        errors = [self.edge_gluing_error(), self.check_cusp_cross_sections()]
        if hasattr(manifold, '_cusp_cross_section_info'):
            errors.append(self.check_against_snappy())
        errors.sort(reverse=True)
        self.error_checks = errors
        assert errors[0] < 1e-7

    def copy(self):
        "Makes a completely seperate copy of self"
        manifold_temp = self.snappy_manifold
        self.snappy_manifold = None
        ans = copy.deepcopy(self)
        self.snappy_manifold = manifold_temp
        ans.snappy_manifold = manifold_temp.copy()
        return ans

    def add_shapes(self):
        "Use HIKMOT to add guaranteed shape intervals." 
        found, shapes = hikmot.verify_hyperbolicity(self.snappy_manifold)
        if not found:
            raise ValueError('HIKMOT could not verify hyperbolic structure')
        for T, z in zip(self.Tetrahedra, shapes):
            add_edge_parameters(T, z)

    def add_cusp_cross_sections(self):
        for cusp in self.Vertices:
            self.add_one_cusp_cross_section(cusp)

    def add_one_cusp_cross_section(self, cusp):
        "Build a cusp cross section as described in Section 3.6 of the paper"
        tet0, vert0 = tets_and_vertices_of_cusp(cusp)[0]
        face0 = FacesAnticlockwiseAroundVertices[vert0][0]
        tet0.horotriangles[vert0] = HoroTriangle(tet0, vert0, face0, hikmot.interval.interval(1))
        active = [(tet0, vert0)]
        while active:
            tet0, vert0 = active.pop()
            for face0 in FacesAnticlockwiseAroundVertices[vert0]:
                tet1, face1 = glued_to(tet0, face0)
                vert1 = tet0.Gluing[face0].image(vert0)
                if tet1.horotriangles[vert1] is None:
                    tet1.horotriangles[vert1] = HoroTriangle(tet1, vert1, face1,
                                tet0.horotriangles[vert0].lengths[face0])
                    active.append( (tet1, vert1) )

    def _get_cusp(self, cusp):
        """
        Helper method so the user can specify a cusp by its index as well
        the actual t3m.Vertex.
        """
        if not isinstance(cusp, t3m.Vertex):
            cusp = self.Vertices[cusp]
        return cusp
            
    def cusp_area(self, cusp):
        cusp = self._get_cusp(cusp)
        area = 0
        for T, V in tets_and_vertices_of_cusp(cusp):
            area += T.horotriangles[V].area
        return area

    def rescale_cusp(self, cusp, scale):
        cusp = self._get_cusp(cusp)
        for T, V in tets_and_vertices_of_cusp(cusp):
            T.horotriangles[V].rescale(scale)
            
    def normalize_cusp(self, cusp):
        """
        Rescale cusp to have area (3/8) sqrt(3), matching SnapPy's conventions.
        """
        cusp = self._get_cusp(cusp)
        # (3/8) sqrt(3) per the kernel's conventions
        target_area = hikmot.interval.interval(0.6495190528383288,
                                               0.6495190528383290)
        area = self.cusp_area(cusp)
        ratio = (target_area/area).sqrt()
        self.rescale_cusp(cusp, ratio)

    def LHS_of_convexity_equations(self):
        """
        For each face in the triangulation, return a quantity which is < 0
        if and only if the corresponding pair of tetrahedra are strictly
        convex.
        """
        ans = []
        for tet0 in self.Tetrahedra:
            for vert0 in ZeroSubsimplices:
                tet1, face1 = glued_to(tet0, comp(vert0))
                ans.append(tet0.tilt(vert0) + tet1.tilt(comp(face1)))
        return ans

    def least_convexity(self):
        """
        Provable upper bound on the largest tilt.  
        """
        return max_sup(self.LHS_of_convexity_equations())
        
    def is_Euclidean_decomposition(self):
        """
        Returns whether the triangulation is the Epstein-Penner
        decomposition for the current choice of cusps.
        """

        tilts = self.LHS_of_convexity_equations()
        if max_sup(tilts) < 0:
            return True
        if max_inf(tilts) > 0:
            return False
        return "unknown"

    def is_canonical_triangulation(self):
        """
        Returns True if this is *provably* the canonical triangulation,
        and False if it is *provably* not, and "unknown" if the tilt
        intervals are ambiguous.

        The optional argument epsilon is only relevant when the manifold
        has two cusps.  It refers to how close the inequalities in equation
        (3.8) of the paper are to being equalities.      
        """
        num_cusps = len(self.Vertices)
        if num_cusps > 1:
            for i in range(num_cusps):
                self.normalize_cusp(i)

        return self.is_Euclidean_decomposition()
                    
    #------- All remaining methods are just sanity and regression checks; safety first! -----------
    
    def check_cusp_cross_sections(self):
        """
        Sanity check: do all pairs of adjacent triangles have the same
        edge lengths?
        """
        errors = []
        for tet0 in self.Tetrahedra:
            for vert0 in ZeroSubsimplices:
                for face0 in FacesAnticlockwiseAroundVertices[vert0]:
                    tet1, face1 = glued_to(tet0, face0)
                    vert1 = tet0.Gluing[face0].image(vert0)
                    side0 = tet0.horotriangles[vert0].lengths[face0]
                    side1 = tet1.horotriangles[vert1].lengths[face1]
                    errors.append(abs(side0-side1))
        return max_sup(errors)

    def check_tilts_against_snappy(self):
        snappy_tilts = self.snappy_manifold._cusp_cross_section_info()[0]
        diffs = []
        for i, tet in enumerate(self.Tetrahedra):
            for v in range(4):
                our_tilt = tet.tilt(ZeroSubsimplices[v])
                snappy_tilt = float(snappy_tilts[i][v])
                diffs.append(abs(our_tilt.mid(our_tilt) - snappy_tilt))
        return diffs

    def check_edge_lengths_against_snappy(self):
        snappy_edges = self.snappy_manifold._cusp_cross_section_info()[1]
        diffs = []
        for i, tet in enumerate(self.Tetrahedra):
            for v in range(4):
                V = ZeroSubsimplices[v]
                for f in range(4):
                    if v != f:
                        F = comp(ZeroSubsimplices[f])
                        our_length = tet.horotriangles[V].lengths[F]
                        snappy_length = float(snappy_edges[i][v][f])
                diffs.append(abs(our_length.mid(our_length)-snappy_length))
        return diffs

    def check_against_snappy(self):
        """
        Compare what we've computed to what SnapPy has internally.
        """
        if not hasattr(self.snappy_manifold, '_cusp_cross_section_info'):
            raise NotImplementedError("Your version of 'snappy' is too old to run this test")

        C = self.copy()
        for cusp in C.Vertices:
            C.normalize_cusp(cusp)
        diffs = C.check_edge_lengths_against_snappy()
        diffs += C.check_tilts_against_snappy()
        return max(diffs)
        
                    
    def edge_gluing_error(self):
        errors = []
        for edge in self.Edges:
            e = 1
            for corner in edge.Corners:
                tet = corner.Tetrahedron
                e = e * tet.edge_params[corner.Subsimplex]
            errors.append(abs(e - 1))
                
        return max_sup(errors)

#--- convenience functions ------

def is_canonical_triangulation(manifold):
    C = CuspedManifold(manifold)
    return C.is_canonical_triangulation()

def least_convexity(manifold):
    C =CuspedManifold(manifold)
    return C.least_convexity()

if __name__ == '__main__':
   import doctest
   doctest.testmod()
